﻿//= OCEAN MODULE ==================================================

// Version 1.0 (dated 19 May 2023)
// This module is an update of a three.js wave generator created in 2015 by Jérémy Bouny (github.com/fft-ocean),
// based on a 2014 js version created by David Li (david.li/waves/) and adapted to three.js by Aleksandr Albert
// This version was created by Phil Crowther <phil@philcrowther.com> on 19 May 2023 and is
// Licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License.
// to the extent that such license is not inconsistent with rights claimed by the above-named authors

// This module should be placed in a location where the module can access the three.module.js 
// (See suggestions at: philcrowther.com/Aviation/Mod_Ocean.html)

import {
	ClampToEdgeWrapping,
	DataTexture,
	FloatType,
	FrontSide,
	LinearFilter,
	LinearMipMapLinearFilter,
	Mesh,
	NearestFilter,
	OrthographicCamera,
	PlaneGeometry,
	RepeatWrapping,
	RGBAFormat,
	Scene,
	ShaderLib,
	ShaderMaterial,
	UniformsUtils,
	Vector2,
	Vector3,
	WebGLRenderTarget
} from "../../../build/three.module.js";

//= OCEAN =======================================================
// Original 2014: Aleksandr Albert (routter.co.tt)
// Modified 2015: Jeremy Bouny (github.com/fft-ocean)
// Modified 2023: Phil Crowther (philcrowther.com)

//= Initialize Ocean ============================================
let Ocean = function (renderer, camera, scene, wav_) {
	// flag used to trigger parameter changes
	this.changed = true;
	this.initial = true;
	// Assign required parameters as object properties
	this.oceanCamera = new OrthographicCamera();
	this.oceanCamera.position.z = 1;
	this.renderer = renderer;
	this.scene = new Scene();
	// Enable necessary extensions
	this.renderer.getContext().getExtension('OES_texture_float');
	this.renderer.getContext().getExtension('OES_texture_float_linear');
	// Load Variables
	this.Siz = wav_.Siz;
	this.windX = wav_.WSp*Math.sin(wav_.WHd*Math.PI/180);
	this.windY = wav_.WSp*Math.cos(wav_.WHd*Math.PI/180);
	this.Chp = wav_.Chp;
	this.Res = wav_.Res;
	//
	this.matrixNeedsUpdate = false;
	
	//= Setup framebuffer pipeline =================================
	let BaseParams = {
		format: RGBAFormat,
		stencilBuffer: false,
		depthBuffer: false,
		premultiplyAlpha: false,
		type: FloatType
	};
	// Set Parameters
	let NearestRepeatParams = JSON.parse(JSON.stringify(BaseParams));
		NearestRepeatParams.minFilter = NearestRepeatParams.magFilter = NearestFilter;
		NearestRepeatParams.wrapS = NearestRepeatParams.wrapT = RepeatWrapping;
	let NearestClampParams = JSON.parse(JSON.stringify(BaseParams));
		NearestClampParams.minFilter = NearestClampParams.magFilter = NearestFilter;
		NearestClampParams.wrapS = NearestClampParams.wrapT = ClampToEdgeWrapping;
	let LinearRepeatParams = JSON.parse(JSON.stringify(BaseParams));
		LinearRepeatParams.minFilter = LinearMipMapLinearFilter;		// ### Add MipMaps
		LinearRepeatParams.generateMipmaps = true;						// ###
		LinearRepeatParams.magFilter = LinearFilter;					// ###
		LinearRepeatParams.wrapS = LinearRepeatParams.wrapT = RepeatWrapping;
	//= Set Render Targets =========================================
	// 8 Buffers of Size = 512x512
	this.initialSpectrumFramebuffer = new WebGLRenderTarget(this.Res, this.Res, NearestRepeatParams);
	this.pingPhaseFramebuffer = new WebGLRenderTarget(this.Res, this.Res, NearestClampParams);
	this.pongPhaseFramebuffer = new WebGLRenderTarget(this.Res, this.Res, NearestClampParams);
	this.spectrumFramebuffer = new WebGLRenderTarget(this.Res, this.Res, NearestClampParams);
	this.pingTransformFramebuffer = new WebGLRenderTarget(this.Res, this.Res, NearestClampParams);
	this.pongTransformFramebuffer = new WebGLRenderTarget(this.Res, this.Res, NearestClampParams);
	this.displacementMapFramebuffer = new WebGLRenderTarget(this.Res, this.Res, LinearRepeatParams);
	this.normalMapFramebuffer = new WebGLRenderTarget(this.Res, this.Res, LinearRepeatParams);
	
	//= Define shaders and constant uniforms =======================
	
	// Vertex Shader - Common
	let fullscreeenVertexShader = ShaderLib["ocean_sim_vertex"];	// Vertex shader  used in all of the simulation steps
	
	// 1 - Fragment Shader - Initial Spectrum
	let initialSpectrumShader = ShaderLib["ocean_initial_spectrum"];
	let initialSpectrumUniforms = UniformsUtils.clone(initialSpectrumShader.uniforms);
	this.materialInitialSpectrum = new ShaderMaterial({
		uniforms: initialSpectrumUniforms,
		vertexShader: fullscreeenVertexShader.vertexShader,	// default vertex shader
		fragmentShader:initialSpectrumShader.fragmentShader
	});
	this.materialInitialSpectrum.uniforms.u_wind = {type: "v2", value: new Vector2()};
	this.materialInitialSpectrum.uniforms.u_grdres = {type: "f", value: this.Res};
	this.materialInitialSpectrum.uniforms.u_grdsiz = {type: "f", value: this.Siz};		// 230412
	this.materialInitialSpectrum.depthTest = false;
	
	// 2 - Fragment Shader - Current Phase
	let phaseShader = ShaderLib["ocean_phase"];
	let phaseUniforms = UniformsUtils.clone(phaseShader.uniforms);
	this.materialPhase = new ShaderMaterial({
		uniforms: phaseUniforms,
		vertexShader: fullscreeenVertexShader.vertexShader,	// default vertex shader
		fragmentShader: phaseShader.fragmentShader
	});
	this.materialPhase.uniforms.u_grdres = {type: "f", value: this.Res};
	this.materialPhase.depthTest = false;
	
	// 3 - Fragment Shader - Current Spectrum
	let spectrumShader = ShaderLib["ocean_spectrum"];
	let spectrumUniforms = UniformsUtils.clone(spectrumShader.uniforms);
	this.materialSpectrum = new ShaderMaterial({
		uniforms: spectrumUniforms,
		vertexShader: fullscreeenVertexShader.vertexShader,
		fragmentShader: spectrumShader.fragmentShader
	});
	this.materialSpectrum.uniforms.u_begFFT = {type: "t", value: null};
	this.materialSpectrum.uniforms.u_grdres = {type: "f", value: this.Res};
	this.materialSpectrum.uniforms.u_choppy.value = this.Chp;
	this.materialSpectrum.depthTest = false;

	// 4 - Fragment Shader - Displacement Map
	// 4A - Horizontal wave vertices used for FFT
	let oceanHorizontalShader = ShaderLib["ocean_subtransform"];	// 2H. Fragment shader used to subtransform the mesh (generates the displacement map)
	let oceanHorizontalUniforms = UniformsUtils.clone(oceanHorizontalShader.uniforms);
	this.materialOceanHorizontal = new ShaderMaterial({
		uniforms: oceanHorizontalUniforms,
		vertexShader: fullscreeenVertexShader.vertexShader,		// default vertex shader
		fragmentShader: "#define HORIZONTAL \n" + oceanHorizontalShader.fragmentShader
	});
	this.materialOceanHorizontal.uniforms.u_transformSize = {type: "f", value: this.Res};
	this.materialOceanHorizontal.uniforms.u_subtransformSize = {type: "f", value: this.Siz};	// 230412
	this.materialOceanHorizontal.uniforms.u_input = {type: "t", value: null};	// u_input.value = ocean_subtransform fragment shader
	this.materialOceanHorizontal.depthTest = false;
	// 4B - Vertical wave vertices used for FFT
	let oceanVerticalShader = ShaderLib["ocean_subtransform"];			// 2V. Fragment shader used to subtransform the mesh (generates the displacement map)
	let oceanVerticalUniforms = UniformsUtils.clone(oceanVerticalShader.uniforms);
	this.materialOceanVertical = new ShaderMaterial({
		uniforms: oceanVerticalUniforms,
		vertexShader: fullscreeenVertexShader.vertexShader,	// default vertex shader
		fragmentShader: oceanVerticalShader.fragmentShader
	});
	this.materialOceanVertical.uniforms.u_transformSize = {type: "f", value: this.Res};
	this.materialOceanVertical.uniforms.u_subtransformSize = {type: "f", value: this.Siz};	// 230412
	this.materialOceanVertical.uniforms.u_input = {type: "t", value: null};		// u_input.value = ocean_subtransform fragment shader
	this.materialOceanVertical.depthTest = false;

	// 6 - Fragment Shader - Normal Map
	let normalShader = ShaderLib["ocean_normals"];
	let normalUniforms = UniformsUtils.clone(normalShader.uniforms);
	this.materialNormal = new ShaderMaterial({
		uniforms: normalUniforms,
		vertexShader: fullscreeenVertexShader.vertexShader,	// default vertex shader
		fragmentShader: normalShader.fragmentShader
	});
	this.materialNormal.uniforms.u_displacementMap = {type: "t", value: null};
	this.materialNormal.uniforms.u_grdres = {type: "f", value: this.Res};
	this.materialNormal.depthTest = false;
	
	// Disable blending to prevent default premultiplied alpha values
	this.materialOceanHorizontal.blending = 0;
	this.materialOceanVertical.blending = 0;
	this.materialInitialSpectrum.blending = 0;
	this.materialPhase.blending = 0;
	this.materialSpectrum.blending = 0;
	this.materialNormal.blending = 0;
	
	// Create the simulation plane
	this.screenQuad = new Mesh(new PlaneGeometry(2, 2));
	this.scene.add(this.screenQuad);	// show screenQuad (no texture?)
		
	// Initialise spectrum data (random phase (0 to 360 degrees) in radians
	this.pingPhase = true;
	let phaseArray = new window.Float32Array(this.Res * this.Res * 4);
	for (let y = 0; y < this.Res; y++) {
		for (let x = 0; x < this.Res; x++) {
			phaseArray[y*this.Res*4 + x*4] = Math.random() * 2.0 * Math.PI;
			phaseArray[y*this.Res*4 + x*4+1] = 0.0;
			phaseArray[y*this.Res*4 + x*4+2] = 0.0;
			phaseArray[y*this.Res*4 + x*4+3] = 0.0;
		}
	}
	this.pingPhaseTexture = new DataTexture(phaseArray, this.Res, this.Res, RGBAFormat);
	this.pingPhaseTexture.minFilter = NearestFilter;
	this.pingPhaseTexture.magFilter = NearestFilter;
	this.pingPhaseTexture.wrapS = ClampToEdgeWrapping;
	this.pingPhaseTexture.wrapT = ClampToEdgeWrapping;
	this.pingPhaseTexture.type = FloatType;
	this.pingPhaseTexture.needsUpdate = true;
	
	// Static Targets
	wav_.Dsp = this.displacementMapFramebuffer.texture;
	wav_.Nrm = this.normalMapFramebuffer.texture;
};

// = OCEAN.RENDER = (called by Main Program) ====================
Ocean.prototype.render = function (wav_) {
	this.scene.overrideMaterial = null;
	
	// 1. Initial Spectrum
	if (this.changed) {
		this.scene.overrideMaterial = this.materialInitialSpectrum;
		this.materialInitialSpectrum.uniforms.u_wind.value.set(this.windX, this.windY);
		this.materialInitialSpectrum.uniforms.u_grdsiz.value = this.Siz;
		this.renderer.setRenderTarget(this.initialSpectrumFramebuffer);
		this.renderer.render(this.scene, this.oceanCamera);
	}
	
	// 2. Current Wave Phase (uses Ping and Pont Buffers)
	this.scene.overrideMaterial = this.materialPhase;
	this.screenQuad.material = this.materialPhase;
	if (this.initial) {
		this.materialPhase.uniforms.u_phases.value = this.pingPhaseTexture;
		this.initial = false;
	}
	else {
		this.materialPhase.uniforms.u_phases.value = this.pingPhase ? this.pingPhaseFramebuffer.texture	: this.pongPhaseFramebuffer.texture;
	}
	this.materialPhase.uniforms.u_deltaTime.value = wav_.DTm;
	this.materialPhase.uniforms.u_grdsiz.value = this.Siz;
	this.renderer.setRenderTarget(this.pingPhase ? this.pongPhaseFramebuffer : this.pingPhaseFramebuffer);
	this.renderer.render(this.scene, this.oceanCamera);
	this.pingPhase = !this.pingPhase;
	
	// 3. renderSpectrum (combination of initial Spectrum and Wave Phases)
	this.scene.overrideMaterial = this.materialSpectrum;
	this.materialSpectrum.uniforms.u_begFFT.value = this.initialSpectrumFramebuffer.texture;
	this.materialSpectrum.uniforms.u_phases.value = this.pingPhase ? this.pingPhaseFramebuffer.texture : this.pongPhaseFramebuffer.texture;
	this.materialSpectrum.uniforms.u_choppy.value = this.Chp;			// 230412
	this.materialSpectrum.uniforms.u_grdsiz.value = this.Siz;
	this.renderer.setRenderTarget(this.spectrumFramebuffer);
	this.renderer.render(this.scene, this.oceanCamera);
	
	// 4. Displacement Map (iterations = 9*2
	let iterations = Math.log2(this.Res)*2; // log2(512) = 9 *2 = 18
	this.scene.overrideMaterial = this.materialOceanHorizontal;
	let subtransformProgram = this.materialOceanHorizontal;
	// Processus 0-N
	// material = materialOceanHorizontal
	// 0 : material(spectrumFramebuffer) > pingTransformFramebuffer
	// i%2==0 : material(pongTransformFramebuffer) > pingTransformFramebuffer
	// i%2==1 : material(pingTransformFramebuffer) > pongTransformFramebuffer
	// i == N/2 : material = materialOceanVertical
	// i%2==0 : material(pongTransformFramebuffer) > pingTransformFramebuffer
	// i%2==1 : material(pingTransformFramebuffer) > pongTransformFramebuffer
	// N-1 : materialOceanVertical(pingTransformFramebuffer / pongTransformFramebuffer) > displacementMapFramebuffer
	let frameBuffer, inputBuffer = 0;
	for (let i = 0; i < iterations-1; i++) {
		if (i === 0) {
			inputBuffer = this.spectrumFramebuffer;
			frameBuffer = this.pingTransformFramebuffer;
		} 
		else if (i%2 === 1) {
			inputBuffer = this.pingTransformFramebuffer;
			frameBuffer = this.pongTransformFramebuffer;
		}
		else {
			inputBuffer = this.pongTransformFramebuffer;
			frameBuffer = this.pingTransformFramebuffer;
		}	
		if (i === iterations/2) {				// switch to vertical at midway point
			subtransformProgram = this.materialOceanVertical;
			this.scene.overrideMaterial = this.materialOceanVertical;
		}
		subtransformProgram.uniforms.u_input.value = inputBuffer.texture;
		subtransformProgram.uniforms.u_subtransformSize.value = Math.pow(2,(i%(iterations/2)+1));
		this.renderer.setRenderTarget(frameBuffer);
		this.renderer.render(this.scene, this.oceanCamera);
	}
	// Final Render to Displacement Map
	inputBuffer = ((iterations%2 === 0)? this.pingTransformFramebuffer : this.pongTransformFramebuffer);
	frameBuffer = this.displacementMapFramebuffer;
	subtransformProgram.uniforms.u_input.value = inputBuffer.texture;
	subtransformProgram.uniforms.u_subtransformSize.value = this.Res;		// 230412
	this.renderer.setRenderTarget(frameBuffer);
	this.renderer.render(this.scene, this.oceanCamera);
	// 5. Normal Map
	this.scene.overrideMaterial = this.materialNormal;
	if (this.changed) this.materialNormal.uniforms.u_grdsiz.value = this.Siz;
	this.materialNormal.uniforms.u_displacementMap.value = this.displacementMapFramebuffer.texture;
	this.renderer.setRenderTarget(this.normalMapFramebuffer);
	this.renderer.render(this.scene, this.oceanCamera);
	
	// Cleanup
	this.scene.overrideMaterial = null;
	this.renderer.setRenderTarget(null);
};

//= FFT OCEAN SHADER ============================================
// Original 2014: David Li (david.li/waves/)
// Modified 2014: Aleksandr Albert (www.routter.co.tt)
// Modified 2015: Jérémy Bouny (github.com/fft-ocean)
// Modified 2023: Phil Crowther (philcrowther.com)
// Description: A deep water ocean shader set based on an implementation of a Tessendorf Waves
// The (modified) set uses 1 common Vertex Shader and 6 Fragment Shaders:
// [1] ocean_initial_spectrum   -> Fragment shader used to set intitial wave frequency at a texel coordinate
// [2] ocean_phase              -> Fragment shader used to set wave phase at a texel coordinate
// [3] ocean_spectrum           -> Fragment shader used to set current wave frequency at a texel coordinate
// [4] ocean_subtransform       -> Fragment shader used to subtransform the mesh (generates the displacement map)
// [5] ocean_normals            -> Fragment shader used to set face normals at a texel coordinate
//
ShaderLib['ocean_sim_vertex'] = {
	varying: {
		"vUV": {type: "v2"}
	},
	vertexShader: [
		'varying vec2 vUV;',
		'void main (void) {',
			'vUV = position.xy * 0.5 + 0.5;',
			'gl_Position = vec4(position, 1.0);',
		'}'
	].join('\n')
};
//= 1. Fragment shader used to set intitial wave frequency at a texel coordinate
// Time Not a Factor
ShaderLib['ocean_initial_spectrum'] = {
	uniforms: {
		"u_wind": {type: "v2", value: new Vector2(10.0, 10.0)},
		"u_grdres": {type: "f", value: null},
		"u_grdsiz": {type: "f", value: null},
	},
	fragmentShader: [
		'const float PI = 3.14159265359;',
		'const float G = 9.81;',
		'const float KM = 370.0;',
		'const float CM = 0.23;',
		'uniform vec2 u_wind;',
		'uniform float u_grdres;',
		'uniform float u_grdsiz;',		
		'float square (float x) {',
			'return x*x;',
		'}',
		'float omega (float k) {',
			'return sqrt(G*k*(1.0+square(k/KM)));',
		'}',
		'float tanH (float x) {',
			'return (1.0-exp(-2.0*x))/(1.0+exp(-2.0*x));',
		'}',
		'void main (void) {',
			'vec2 coordinates = gl_FragCoord.xy-0.5;',			
			'float n = (coordinates.x < u_grdres * 0.5) ? coordinates.x : coordinates.x - u_grdres;',
			'float m = (coordinates.y < u_grdres * 0.5) ? coordinates.y : coordinates.y - u_grdres;',		
			'vec2 K = (2.0*PI*vec2(n,m))/u_grdsiz;',
			'float k = length(K);',		
			'float l_wind = length(u_wind);',
			'float Omega = 0.84;',
			'float kp = G*square(Omega/l_wind);',
			'float c = omega(k)/k;',
			'float cp = omega(kp)/kp;',
			'float Lpm = exp(-1.25*square(kp/k));',
			'float gamma = 1.7;',
			'float sigma = 0.08*(1.0+4.0*pow(Omega, -3.0));',
			'float Gamma = exp(-square(sqrt(k/kp)-1.0)/2.0*square(sigma));',
			'float Jp = pow(gamma, Gamma);',
			'float Fp = Lpm*Jp*exp(-Omega/sqrt(10.0)*(sqrt(k/kp)-1.0));',
			'float alphap = 0.006*sqrt(Omega);',
			'float Bl = 0.5*alphap*cp/c*Fp;',
			'float z0 = 0.000037 * square(l_wind)/G*pow(l_wind/cp, 0.9);',
			'float uStar = 0.41*l_wind/log(10.0/z0);',
			'float alpham = 0.01*((uStar<CM) ? (1.0+log(uStar/CM)) : (1.0+3.0*log(uStar/CM)));',
			'float Fm = exp(-0.25*square(k/KM-1.0));',
			'float Bh = 0.5*alpham*CM/c*Fm*Lpm;',
			'float a0 = log(2.0)/4.0;',
			'float am = 0.13*uStar/CM;',
			'float Delta = tanH(a0+4.0*pow(c/cp, 2.5)+am*pow(CM/c, 2.5));',
			'float cosPhi = dot(normalize(u_wind), normalize(K));',
			'float S = (1.0/(2.0*PI))*pow(k,-4.0)*(Bl+Bh)*(1.0+Delta*(2.0*cosPhi*cosPhi-1.0));',
			'float dk = 2.0*PI/u_grdsiz;',
			'float h = sqrt(S/2.0)*dk;',
			'if (K.x == 0.0 && K.y == 0.0) {h = 0.0;}', 	//no DC term
			'gl_FragColor = vec4(h, 0.0, 0.0, 0.0);',
		'}'
	].join('\n')
};
//= 2. Fragment Shader - Current Phase
ShaderLib['ocean_phase'] = {
	uniforms: {
		"u_phases": {type: "t", value: null},
		"u_deltaTime": {type: "f", value: null},
		"u_grdres": {type: "f", value: null},
		"u_grdsiz": {type: "f", value: null},
	},
	varying: {
		"vUV": {type: "v2"}
	},
	fragmentShader: [
		'const float PI = 3.14159265359;',
		'const float G = 9.81;',
		'const float KM = 370.0;',
		'varying vec2 vUV;',
		'uniform sampler2D u_phases;',
		'uniform float u_deltaTime;',
		'uniform float u_grdres;',
		'uniform float u_grdsiz;',
		'float omega (float k) {',
			'return sqrt(G*k*(1.0+k*k/KM*KM));',
		'}',
		'void main (void) {',
			'vec2 coordinates = gl_FragCoord.xy-0.5;',
			'float n = (coordinates.x < u_grdres*0.5) ? coordinates.x : coordinates.x - u_grdres;',
			'float m = (coordinates.y < u_grdres*0.5) ? coordinates.y : coordinates.y - u_grdres;',
			'vec2 waveVector = (2.0*PI*vec2(n,m))/u_grdsiz;',
			'float phase = texture2D(u_phases, vUV).r;',
			'float deltaPhase = omega(length(waveVector))*u_deltaTime;',
			'phase = mod(phase+deltaPhase, 2.0*PI);',	
			'gl_FragColor = vec4(phase, 0.0, 0.0, 0.0);',
		'}'
	].join('\n')
};
//= 3. Fragment Shader - Current Spectrum
ShaderLib['ocean_spectrum'] = {
	uniforms: {
		"u_grdsiz": {type: "f", value: null},
		"u_grdres": {type: "f", value: null},
		"u_choppy": {type: "f", value: null},
		"u_phases": {type: "t", value: null},
		"u_begFFT": {type: "t", value: null},
	},
	varying: {
		"vUV": {type: "v2"}
	},
	fragmentShader: [
		'const float PI = 3.14159265359;',
		'const float G = 9.81;',
		'const float KM = 370.0;',
		'varying vec2 vUV;',
		'uniform float u_grdsiz;',
		'uniform float u_grdres;',
		'uniform float u_choppy;',
		'uniform sampler2D u_phases;',
		'uniform sampler2D u_begFFT;',
		'vec2 multiplyComplex (vec2 a, vec2 b) {',
			'return vec2(a[0]*b[0]-a[1]*b[1], a[1]*b[0]+a[0]*b[1]);',
		'}',
		'vec2 multiplyByI (vec2 z) {',
			'return vec2(-z[1], z[0]);',
		'}',
		'float omega (float k) {',
			'return sqrt(G*k*(1.0+k*k/KM*KM));',
		'}',
		'void main (void) {',
			'vec2 coordinates = gl_FragCoord.xy - 0.5;',
			'float n = (coordinates.x < u_grdres * 0.5) ? coordinates.x : coordinates.x - u_grdres;',
			'float m = (coordinates.y < u_grdres * 0.5) ? coordinates.y : coordinates.y - u_grdres;',
			'vec2 waveVector = (2.0 * PI * vec2(n, m)) / u_grdsiz;',
			'float phase = texture2D(u_phases, vUV).r;',
			'vec2 phaseVector = vec2(cos(phase), sin(phase));',
			'vec2 h0 = texture2D(u_begFFT, vUV).rg;',
			'vec2 h0Star = texture2D(u_begFFT, vec2(1.0 - vUV + 1.0 / u_grdres)).rg;',
			'h0Star.y *= -1.0;',
			'vec2 h = multiplyComplex(h0, phaseVector) + multiplyComplex(h0Star, vec2(phaseVector.x, -phaseVector.y));',
			'vec2 hX = -multiplyByI(h * (waveVector.x / length(waveVector))) * u_choppy;',
			'vec2 hZ = -multiplyByI(h * (waveVector.y / length(waveVector))) * u_choppy;',
			//no DC term
			'if (waveVector.x == 0.0 && waveVector.y == 0.0) {',
				'h = vec2(0.0);',
				'hX = vec2(0.0);',
				'hZ = vec2(0.0);',
			'}',	
			'gl_FragColor = vec4(hX + multiplyByI(h), hZ);',
		'}'
	].join('\n')
};
//= 4. Fragment Shader - Displacement Map
// output textures = materialOceanHorizontal and materialOceanVertical
ShaderLib['ocean_subtransform'] = {
	uniforms: {
		"u_input": {type: "t", value: null},
		"u_transformSize": {type: "f", value: null},
		"u_subtransformSize": {type: "f", value: null}
	},
	varying: {
		"vUV": {type: "v2"}
	},
	fragmentShader: [
		'const float PI = 3.14159265359;',
		'uniform sampler2D u_input;',
		'uniform float u_transformSize;',
		'uniform float u_subtransformSize;',
		'varying vec2 vUV;',	
		'vec2 multiplyComplex (vec2 a, vec2 b) {',
			'return vec2(a[0]*b[0]-a[1]*b[1], a[1]*b[0]+a[0]*b[1]);',
		'}',
		'void main (void) {',
			'#ifdef HORIZONTAL',
				'float index = vUV.x * u_transformSize - 0.5;',		// horizontal
			'#else',
				'float index = vUV.y * u_transformSize - 0.5;',		// vertical
			'#endif',
			'float evenIndex = floor(index / u_subtransformSize) * (u_subtransformSize * 0.5) + mod(index, u_subtransformSize * 0.5);',
			'#ifdef HORIZONTAL',
				'vec4 even = texture2D(u_input, vec2(evenIndex+0.5, gl_FragCoord.y)/u_transformSize).rgba;',
				'vec4 odd = texture2D(u_input, vec2(evenIndex + u_transformSize*0.5+0.5, gl_FragCoord.y)/u_transformSize).rgba;',
			'#else',
				'vec4 even = texture2D(u_input, vec2(gl_FragCoord.x, evenIndex+0.5)/u_transformSize).rgba;',
				'vec4 odd = texture2D(u_input, vec2(gl_FragCoord.x, evenIndex+u_transformSize*0.5+0.5)/u_transformSize).rgba;',
			'#endif',
			'float twiddleArgument = -2.0*PI*(index/u_subtransformSize);',
			'vec2 twiddle = vec2(cos(twiddleArgument), sin(twiddleArgument));',
			'vec2 outputA = even.xy + multiplyComplex(twiddle, odd.xy);',
			'vec2 outputB = even.zw + multiplyComplex(twiddle, odd.zw);',
			'gl_FragColor = vec4(outputA,outputB);',
		'}'
	].join('\n')
};
//= 5. Fragment Shader - Normal Map
// Note: The original program did not generate a "plug and play" Normal Map - modifications were required
ShaderLib['ocean_normals'] = {
	uniforms: {
		"u_displacementMap": {type: "t", value: null},
		"u_grdres": {type: "f", value: null},
		"u_grdsiz": {type: "f", value: null},
	},
	varying: {
		"vUV": {type: "v2"}
	},
	fragmentShader: [
		'varying vec2 vUV;',		
		'uniform sampler2D u_displacementMap;',
		'uniform float u_grdres;',
		'uniform float u_grdsiz;',
		'void main (void) {',
			'float texel = 1.0/u_grdres;',
			'float texelSize = u_grdsiz/u_grdres;',
			'vec3 ctr = texture2D(u_displacementMap, vUV).rgb;',
			'vec3 rgt = vec3(texelSize, 0.0, 0.0) + texture2D(u_displacementMap, vUV + vec2(texel, 0.0)).rgb - ctr;',
			'vec3 lft = vec3(-texelSize, 0.0, 0.0) + texture2D(u_displacementMap, vUV + vec2(-texel, 0.0)).rgb - ctr;',
			'vec3 top = vec3(0.0, 0.0, -texelSize) + texture2D(u_displacementMap, vUV + vec2(0.0, -texel)).rgb - ctr;',
			'vec3 bot = vec3(0.0, 0.0, texelSize) + texture2D(u_displacementMap, vUV + vec2(0.0, texel)).rgb - ctr;',
			'vec3 topRgt = cross(rgt, top);',
			'vec3 topLft = cross(top, lft);',
			'vec3 botLft = cross(lft, bot);',
			'vec3 botRgt = cross(bot, rgt);',
			'vec3 nrm3 = vec3(normalize(topRgt + topLft + botLft + botRgt));',
			'vec3 tmp2 = nrm3;',
			'nrm3.b = tmp2.g;',
			'nrm3.g = tmp2.b;',
			'gl_FragColor = vec4(nrm3*0.5+0.5, 1.0);',
		'}'
	].join('\n')
};

/* Change Log =================================================*/
// 230519: Version 1.0 

export {Ocean};
